#ifndef __READ_DATA_C__
#define __READ_DATA_C__
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "ascii_util.h"
#include "station.h"

#define BUFSIZE 10
#define FILE_NAME_LENGTH 128 
#define COMMENT_LENGTH 72
#define DAQKEY 50
#define FADC_MAX_MODULE 6                    /* # of f_adcs */
#define SADC_MAX_MODULE 2                    /* # of s_adcs */
#define FADC_BLOCK_LENGTH 8192
#define SADC_BLOCK_LENGTH 16

void main(int argc, char **argv)
{
    FILE *filed = NULL, *wfiled = NULL;
    int input_run_number;
    char input_filename[FILE_NAME_LENGTH];
    char output_filename[FILE_NAME_LENGTH];

    unsigned int run_number;
    unsigned int number_of_events;
    short int fadc_module, sadc_module;
    unsigned short int number_of_station_f, number_of_station_s;
    unsigned int real_events;
    time_t start_time, stop_time;
    char start_comment[COMMENT_LENGTH];
    char end_comment[COMMENT_LENGTH];
    short int thred[2];

    int naf;
    div_t d_naf;
    int j, x, y, n, a, i;
    int nword;
    int FORMAT_ID = 1;
    int header_length = 154;
    int trailer_length = 44;
    short int urh_length = 6;
    short int hard_map_length = 97;
    short int thred_length = 13;
    short int ev;
    short int type;
    short int nhit;
    short int w_start, w_stop;
    unsigned short int chamber, crt, hodo, ch, c_wire, h_wire;
    unsigned short int fadc[4096*12 + 36], time[4096*12 + 36];
    short int stn;
    unsigned short int snac = 0;
    unsigned short int sadc_channel = 12;
    short int reset_flag = 0;
    int sow = 0, tsow = 0, nbe = 0, nre = 0, nne = 0;

    /*
     * =-=-=-=-=- Input run number
     */
    printf("argc = %d\n",argc);

    if (argc <= 1) {
        printf("Please input Run number[integer]: ");
        scanf("%d", &input_run_number);
    } else {
        input_run_number = atoi(argv[1]);
    }
    sprintf(input_filename, "Run%5.5d.raw", input_run_number);
    sprintf(output_filename, "Run%5.5d.dst", input_run_number);

    if ((filed = fopen(input_filename, "r")) == NULL)  {
        exit(2);
    }
    if ((wfiled = fopen(output_filename, "w")) == NULL) {
        exit(2);
    }

    /*
     * =-=-=-=-=- Header part (106words)
     */
/* 
 * ------------------------------------------------------------------------
 * Raw Data Format for Two-Track Separation Experiment 1999
 * ------------------------------------------------------------------------
 * (Update Record)
 *   1999/06/03 K.Fujii	Version 1. 
 * 
 * [1] Universal Record Header Format
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|	        				| 0
 * 	+--+--+--+--+-  No. of Words (NW)  -+--+--+--+--+
 * 	|		(including itself)		| 1
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|		Format ID			| 2
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| Record Type = (1,2,10,<0) = (BR,ER,ADC,error)	| 3
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|		Run Number			| 4
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|		Event Number			| 5
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * [2] Begin Run
 * 
 * [2-1] URH
 * 
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|						| 0
 * 	|	    Fixed Length (6 words)		|
 * 	|	    Universal Record Header		|
 * 	|						|
 * 	|		Record Type = 1			|
 * 	|						| 5
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * [2-2] Time
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|	        				| 0 + 6
 * 	+--+--+--+--+-  Time               -+--+--+--+--+
 * 	|						| 1 + 6
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *
 * [2-3] Hardware Map
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|  # words for hardware map (nhw)		| 0 + (6+2)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|	Crate #	     |   Station #  | Channel # | 1 + (6+2)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| ID  |  Detector Channel Number		|
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|			 			|
 * 	:			:			:
 * 	: 	ID = 0 : chamber			:
 *	:    	   = 1 : old hodoscope			:
 *	:    	   = 2 : new hodoscope			:
 * 	:			:			:
 * 	|						|
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|	Crate #	     |   Station #  | Channel # |
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| ID  |  Detector Channel Number		| (nhw-1) + (6+2)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * [2-4] Threshold Data for FADC's
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|  # words for threshold data (ntw)		| 0 + (6+2+nhw)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| 	Threshold value for wire #1		| 1 + (6+2+nhw)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|						|
 * 	:			:			:
 * 	:			:			:
 * 	:			:			:
 * 	|						|
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| 	Threshold value for wire #(ntw-1)	| (ntw-1) + (6+2+nhw)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * [2-5] Begin Run Comments
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| 	Begin run comments			| 0 + (6+2+nhw+ntw)
 * 	:			:			:
 * 	:			:			:
 * 	|						| 35 + (6+2+nhw+ntw)
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+ 
 *							  NW = 6+2+nhw+ntw+36
 *
 * 
 * [3] ADC Data (Record Type = 10)
 * 
 * [3-1] URH
 * 
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|						| 0
 * 	|	    Fixed Length (6 words)		|
 * 	|	    Universal Record Header		|
 * 	|						|
 * 	|		Record Type = 10		|
 * 	|						| 5
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * [3-2] ADC Data
 * 
 * 	 15       12	       8	   4	       0
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| Station #    |	ADC value (0-255)	| 6
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| Channel # |	Time clock count (0-4095)	|	
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	|						|
 * 	:			:			:
 * 	:			:			:
 * 	:			:			:
 * 	|						|
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| Station #    |	ADC value (0-255)	|	
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 	| Channel # |	Time clock count (0-4095)	| (NW-1)	
 * 	+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 * 
 * 
 * 
 * 
 * 
 * 
 */ 

    type = 1;
    fread(start_comment, sizeof(start_comment), 1, filed);
    fread((char *) &start_time,  sizeof(time_t), 1, filed);
    fread((char *) &run_number,  sizeof(unsigned int), 1, filed);
    fread((char *) &fadc_module, sizeof(short int), 1, filed);
    fread((char *) &sadc_module, sizeof(short int), 1, filed);
    fread((char *) &number_of_events, sizeof(unsigned int), 1, filed);

    hard_map_length = 1 + 2*(2*fadc_module + sadc_channel*sadc_module);
    thred_length    = 1 + 2*fadc_module;
    header_length   = urh_length + 2 + hard_map_length + thred_length
                      + COMMENT_LENGTH/2;

    fwrite((char *) &header_length, sizeof(int), 1, wfiled);
    fwrite((char *) &FORMAT_ID, sizeof(short int), 1, wfiled);
    fwrite((char *) &type,       sizeof(short int), 1, wfiled);
    fwrite((char *) &run_number, sizeof(short int), 1, wfiled);
    fwrite((char *) &number_of_events, sizeof(short int), 1, wfiled);
    fwrite((char *) &start_time, sizeof(int), 1, wfiled);

    fwrite((char *) &hard_map_length, sizeof(short int), 1, wfiled);
    chamber = 0;
    c_wire = 0;
    crt = 0;
    crt = crt << 9;
    chamber = chamber << 14;
    for(n = 0; n < fadc_module; n++) {
        number_of_station_f = fadc_station[n];
        for(ch = 0; ch < 2; ch++){
            c_wire++;
            number_of_station_f = number_of_station_f << 4;
            snac = ( crt | number_of_station_f | ch );
            fwrite((char *) &snac, sizeof(short int), 1, wfiled);
            snac = ( chamber | c_wire );
            fwrite((char *) &snac, sizeof(short int), 1, wfiled);
        }
    }

    crt = 0;
    crt = crt << 9;
    n = -1;
    hodo = 1;
    for(ch =0; ch < sadc_channel*sadc_module; ch++){
        if (ch > 15) hodo = 2;
        h_wire = ch + 1;
        if (ch % sadc_channel == 0) n++;
        number_of_station_s = sadc_station[n]; 
        number_of_station_s = number_of_station_s << 5;
        snac = ( crt | number_of_station_s | (ch % sadc_channel) );
        fwrite((char *) &snac,sizeof(short int), 1, wfiled);
        hodo = ( hodo << 14 );
        snac = ( hodo | h_wire );
        fwrite((char *) &snac, sizeof(short int), 1, wfiled);
    }    
    
    printf("run_number:       %d\n", run_number);
    printf("number_of_events: %d\n", number_of_events);

    /* read threshold data */
    fwrite((char *) &thred_length, sizeof(short int), 1, wfiled);
    for (j = 0; j < fadc_module; j++) {
        fread((char *) &thred[0], sizeof(int), 1, filed);
        fread((char *) &thred[1], sizeof(int), 1, filed);
#ifdef DUMP
        printf("threshold: %d %d\n", thred[0], thred[1]);
#endif
        fwrite((char *) &thred[0], sizeof(unsigned short int), 1, wfiled);
        fwrite((char *) &thred[1], sizeof(unsigned short int), 1, wfiled);
    }
    fwrite(start_comment, sizeof(start_comment), 1, wfiled);


    /* header information */
    ev = 1;
    sow = header_length;
    tsow = tsow + sow;
    printf("run %d, start, type %d, word %d\n", run_number, type, sow);

    /*
     * =-=-=-=-=- Event data part
     */
    while(1) {
        nword = 0;
        for (naf = 0; naf < fadc_module * 2; naf++) {
            d_naf = div(naf, 2);
            n = d_naf.quot; a = d_naf.rem;

            number_of_station_f = fadc_station[n];

            /* read station tag number */
            fread((char *) &stn, sizeof(short int), 1, filed);
            if (stn < 0)
                goto end;
            /* read # of hit */
            fread((char *) &nhit, sizeof(short int), 1, filed);
#ifdef DUMP
            printf("event %d FADC %d ch %d hit %d\n", ev , n, a, nhit);
#endif

            if (nhit) {
                for(x = 0; x < nhit ; x++) {
                    /* read start/stop time */
                    fread((char *) &w_start, sizeof(short int), 1, filed);
                    fread((char *) &w_stop, sizeof(short int), 1, filed);
#ifdef DUMP
                    printf("write from %d to %d\n", w_start, w_stop);
#endif

                    /* read adc channel */
                    for (y = w_start; y < w_stop + 1; y++) {
                        fread((char *) &fadc[nword], sizeof(short int), 1, filed);
#ifdef DUMP
                        printf("FADC %d %d\n", y, fadc[nword]);
#endif
                        snac = number_of_station_f << 11;
                        fadc[nword] = snac | fadc[nword];
                        snac = a << 12;
                        time[nword++] =  snac | y;

                    } /* end of read adc channel */
                }
            } /* end of all adc channel reading for 1-channel */
        } /* end of FADC loop */

        n = -1;
        for (x = 0; x < sadc_module * sadc_channel; x++) { /* Read ADC data */
            fread((char *) &fadc[nword], sizeof(short int), 1, filed);
            if ( x % sadc_channel == 0 ) n++;
            number_of_station_s = sadc_station[n];
            snac = number_of_station_s << 11;
            fadc[nword] = snac | fadc[nword];
            snac = (x % sadc_channel) << 12;
            time[nword++] = snac;
            
            
#ifdef DUMP
            printf("SADC %d %d\n", x, fadc[nword-1]);
#endif
        } /* end ADC loop */
            
        i = urh_length + nword * 2 ;

        fread((char *) &reset_flag, sizeof(short int), 1, filed);
        if (stn != (fadc_module * 2 - 1)) {
            printf("Wrong event stn=%d\n", stn);
            type = -1;
            nbe++;
        } else if (reset_flag != 0) {
            printf("Reset event flag=%d\n", reset_flag);
            type = -10;
            nre++;
        } else {
            type = 10;
            nne++;
        }
        
        /* write data */
        fwrite((char *) &i, sizeof(int), 1, wfiled);
        fwrite((char *) &FORMAT_ID, sizeof(short int), 1, wfiled);
        fwrite((char *) &type, sizeof(short int), 1, wfiled);
        fwrite((char *) &run_number, sizeof(short int), 1, wfiled);
        fwrite((char *) &ev, sizeof(short int), 1, wfiled);

        for (j = 0; j < nword; j++) {
            fwrite((char *) &fadc[j], sizeof(short int), 1, wfiled);
            fwrite((char *) &time[j], sizeof(short int), 1, wfiled);
        }
           

         /* data information */
        ev++;
        sow = i;
        tsow = tsow + sow;
        printf("run %d, event %d, type %d, word %d\n", run_number, ev, type, sow);

    } /* end of while loop */
    
end:

    /*
     * =-=-=-=-=- end comment part
     */
    type = 2;
    fread((char *)&real_events, sizeof(unsigned int), 1, filed);
    fread((char *)&stop_time, sizeof(time_t), 1, filed);
    fread(end_comment, sizeof(end_comment), 1, filed);

    fwrite((char *) &trailer_length,  sizeof(int), 1, wfiled);
    fwrite((char *) &FORMAT_ID, sizeof(short int), 1, wfiled);
    fwrite((char *) &type, sizeof(short int), 1, wfiled);
    fwrite((char *) &run_number, sizeof(short int), 1, wfiled);
    fwrite((char *) &real_events, sizeof(short int), 1, wfiled);
    fwrite((char *) &stop_time, sizeof(int), 1, wfiled);
    fwrite(end_comment, sizeof(end_comment), 1, wfiled);

    printf("real_events%d\n", ev);
    printf("stop_time %s\n",time2asc(stop_time));
    printf("%s\n", end_comment);

    /* data information */
    sow = header_length;
    tsow = tsow + sow;
    printf("run %d, end, type %d, word %d\n", run_number, type, sow);

    printf("Run %d, event %d, normal %d, bad %d, reset %d, word %d(%d)\n",
           run_number, ev, nne, nbe, nre, tsow, tsow*16);

    /* file close */
    fclose(filed);
}
